use std::{borrow::Cow, collections::HashMap, fmt::Debug, io::BufWriter, primitive};

use petgraph::{
    adj::EdgeIndex,
    algo::{toposort, DfsSpace},
    data::{Build, DataMap},
    graph::DiGraph,
    graphmap::DiGraphMap,
    prelude::NodeIndex,
    visit::NodeIndexable,
};
use quick_xml::{events::attributes::Attribute, name::QName, ElementWriter};

pub mod length;
pub mod nodes;

use crate::types::nodes::primitives::color_matrix::{ColorMatrix, ColorMatrixType};

use self::{
    length::{Coordinate, Length},
    nodes::{
        primitives::{FePrimitive, WriteElement},
        Node,
    },
};

#[derive(Debug)]
pub struct Filter<'a> {
    pub graph: DiGraph<Node, Edge<'a>>,
    source_graphic_idx: NodeIndex,
}

impl Filter<'_> {
    pub fn new() -> Self {
        let mut graph = DiGraph::new();
        let source_graphic_idx = graph.add_node(Node::StdInput(
            nodes::standard_input::StandardInput::SourceGraphic,
        ));

        Self {
            graph,
            source_graphic_idx,
        }
    }

    pub fn add_node(&mut self, node: Node) -> NodeIndex {
        self.graph.add_node(node)
    }

    pub fn source_graphic(&self) -> NodeIndex {
        self.source_graphic_idx
    }

    pub fn to_svg(&self) -> String {
        let mut result = Vec::new();
        let mut writer = quick_xml::Writer::new(&mut result);

        let mut dfs_space = DfsSpace::new(&self.graph);
        let sorted = toposort(&self.graph, Some(&mut dfs_space)).expect("No cycles! Bad user!");

        sorted
            .into_iter()
            .filter_map(|node_idx| {
                let node = self
                    .graph
                    .node_weight(node_idx)
                    .expect("toposorting will not return invalid indices");

                if let Node::Primitive {
                    primitive,
                    common_attrs,
                } = node
                {
                    Some((node_idx, primitive, common_attrs))
                } else {
                    None
                }
            })
            .fold(
                &mut writer,
                |writer, (node_idx, primitive, common_attrs)| {
                    let el_writer = primitive.element_writer(writer);
                    create_input_elements(
                        &self.graph,
                        el_writer,
                        node_idx,
                        self.graph
                            .node_weight(node_idx)
                            .expect("cannot get invalid node_idx from toposort")
                            .input_count(),
                    )
                    .write_empty()
                    .expect("should write successfully")
                },
            );

        String::from_utf8_lossy(&result).to_string()
    }
}

impl Default for Filter<'_> {
    fn default() -> Self {
        Self::new()
    }
}

#[derive(Debug, Clone)]
pub struct Edge<'a> {
    pub edge_type: EdgeType<'a>,
    /// the index of the `in` attribute on the target element
    /// if None, just `in`
    pub in_idx: Option<u8>,
}

impl Edge<'_> {
    pub fn new() -> Self {
        Self {
            edge_type: EdgeType::Unnamed,
            in_idx: None,
        }
    }

    pub fn in_idx(idx: u8) -> Self {
        Self {
            edge_type: EdgeType::Unnamed,
            in_idx: Some(idx),
        }
    }
}

impl Default for Edge<'_> {
    fn default() -> Self {
        Self::new()
    }
}

#[derive(Debug, Clone)]
pub enum EdgeType<'a> {
    Named(&'a str),
    /// For standard inputs such as SourceGraphic etc., which we'll just be representing as nodes for simplicity
    Unnamed,
}
fn create_input_elements<'a>(
    g: &'a DiGraph<Node, Edge<'_>>,
    mut el_writer: ElementWriter<'a, &'a mut Vec<u8>>,
    node_idx: NodeIndex,
    input_count: u8,
) -> ElementWriter<'a, &'a mut Vec<u8>> {
    let inputs = g
        .neighbors_directed(node_idx, petgraph::Direction::Incoming)
        .collect::<Vec<NodeIndex>>();

    if inputs.len() != input_count as usize {
        // TODO: better error handling
        panic!("input couns didnt match");
    }

    inputs
        .into_iter()
        .enumerate()
        .fold(el_writer, |el_writer, (i, incoming_idx)| {
            let incoming_node = g.node_weight(incoming_idx).expect("cannot fail here");

            // find incoming edge and get weight
            let Edge { edge_type, in_idx } = g
                .edge_weight(
                    g.find_edge(incoming_idx, node_idx)
                        .expect("there should always be an edge"),
                )
                .expect("once again, should always exist");

            let in_attr_name = match in_idx {
                None | Some(0) => "in".to_owned(),
                Some(n) => format!("in{}", n + 1),
            };

            let v = match incoming_node {
                Node::StdInput(std_in) => format!("{std_in:?}"),
                Node::Primitive {
                    primitive,
                    common_attrs,
                } => todo!(),
            };

            el_writer.with_attribute((in_attr_name.as_str(), v.as_str()))
        })
}
